using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Reflection;
using System.Threading;
using HarmonyLib;
using RDTools;
using SharpHook;
using UnityEngine;

// ReSharper disable UnusedType.Local

namespace ADOFAIAsyncInput
{
    internal static class Patches
    {
        public static Thread Thread;

        private static FieldInfo _hookField = AccessTools.Field(typeof(AsyncInputManager), "hook");

        internal static void StartThread()
        {
            if (Thread != null)
            {
                return;
            }

            Thread = new Thread(() =>
            {
                try
                {
                    AsyncInputNative.RunHook(data =>
                    {
                        var keyCode = (ushort)data.key;

                        switch (data.event_type)
                        {
                            case AsyncInputNative.EventType.Press:
                                AsyncInputManager.keyQueue.Enqueue(new KeyEvent(DateTime.Now.Ticks, keyCode, true));
                                break;
                            case AsyncInputNative.EventType.Release:
                                AsyncInputManager.keyQueue.Enqueue(new KeyEvent(DateTime.Now.Ticks, keyCode, false));
                                break;
                        }
                    });
                }
                catch (Exception e)
                {
                    AsyncInputMod.Entry.Logger.Error($"{e}");
                    throw;
                }
            });

            Thread.Start();
        }

        [HarmonyPatch(typeof(scrController), "UpdateInput")]
        private static class ControllerUpdateInput
        {
            private static MethodBase ProcessKeyInputs = AccessTools.Method(typeof(scrController), "ProcessKeyInputs");

            private static bool Prefix(scrController __instance, ref bool ___validKeyWasReleased)
            {
                var controller = __instance;

                ___validKeyWasReleased = false;

                AsyncInputManager.keyDownMask.Clear();
                AsyncInputManager.keyUpMask.Clear();
                AsyncInputManager.frameDependentKeyDownMask.Clear();
                AsyncInputManager.frameDependentKeyUpMask.Clear();

                long tick = 0L;

                while (AsyncInputManager.keyQueue.TryDequeue(out var result))
                {
                    ProcessKeyInputs.Invoke(controller, new object[]
                    {
                        new List<KeyCode>
                        {
                            (KeyCode)result.keyCode
                        },
                        result.tick
                    });
                    AsyncInputManager.keyDownMask.Clear();
                    AsyncInputManager.keyUpMask.Clear();

                    tick = result.tick;
                }
                
                ProcessKeyInputs.Invoke(controller, new object[]
                {
                    new List<KeyCode>(),
                    tick
                });

                return false;
            }
        }

#if DEBUG
        [HarmonyPatch(typeof(RDBaseDll), "printem")]
        private static class RdBaseDllPrint
        {
            private static bool Prefix(object o)
            {
                MethodBase method = new StackTrace().GetFrame(2).GetMethod();
                if (method.DeclaringType != null)
                    AsyncInputMod.Entry.Logger.Log(
                        $"{(object)method.DeclaringType.Name}.{(object)method.Name}() -> {o}");
                return false;
            }
        }
#endif

        [HarmonyPatch(typeof(AsyncInputManager), "isActive", MethodType.Getter)]
        private static class AsyncInputManagerIsActive
        {
            private static bool Prefix(ref bool __result)
            {
                __result = Thread != null;

                return false;
            }
        }

        [HarmonyPatch(typeof(AsyncInputManager), "ToggleHook")]
        private static class AsyncInputManagerToggleHook
        {
            private static bool Prefix(bool active)
            {
                IGlobalHook hook = (IGlobalHook)_hookField.GetValue(null);

                if (hook != null)
                {
                    AsyncInputMod.Entry.Logger.Log("Stopping running hook");
                    hook.Dispose();
                    Thread hookThread = AsyncInputManager.hookThread;
                    if ((hookThread != null ? (hookThread.IsAlive ? 1 : 0) : 0) != 0)
                        AsyncInputManager.hookThread.Abort();
                }

                if (active)
                {
                    AsyncInputManager.keyMask.Clear();
                    AsyncInputManager.keyQueue.Clear();
                    AsyncInputManager.keyDownMask.Clear();
                    AsyncInputManager.keyUpMask.Clear();
                    long num;
                    AsyncInputManager.offsetTick = num = 0L;
                    AsyncInputManager.prevFrameTick = num;
                    AsyncInputManager.targetSongTick = num;
                    AsyncInputManager.currFrameTick = num;
                    AsyncInputManager.offsetTickUpdated = false;
                    AsyncInputManager.dspTime = AsyncInputManager.dspTimeSong = 0.0;
                    AsyncInputManager.lastReportedDspTime = AsyncInputManager.previousFrameTime = 0.0;

                    StartThread();
                    AsyncInputMod.Entry.Logger.Log("Started Hook");
                    RDInput.OnAsyncInputToggle(true);
                }
                else
                {
                    if (Thread == null) return false;
                    // AsyncInputNative.Stop();
                    Thread.Abort();
                    AsyncInputMod.Entry.Logger.Log("Stopped Hook");
                    Thread = null;
                    RDInput.OnAsyncInputToggle(false);
                }

                return false;
            }
        }
    }
}